import * as elements from '@yellicode/elements';
import { CodeWriter, CodeWriterUtility } from '@yellicode/core';
import { JavaTypeNameProvider } from './java-type-name-provider';
import { JavaCommentWriter } from './comment-writer';
import { XmlDocUtility } from './xml-doc-utility';
import { DefinitionBuilder } from './definition-builder';
/**
 * A CodeWriter for writing Java code from code generation templates. This writer can write classes, interfaces, structs and enumerations and also
 * contains functions for writing namespace blocks and using directives. The JavaWriter is compatible with Yellicode models but can also work
 * independently.
 */
export class JavaWriter extends CodeWriter {
    /**
     * Constructor. Creates a new JavaWriter instance using the TextWriter and options provided.
     * @param writer The template's current TextWriter.
     * @param options Optional: the global options for this writer.
     */
    constructor(writer, options) {
        super(writer);
        if (!options)
            options = {};
        this.typeNameProvider =
            options.typeNameProvider || new JavaTypeNameProvider();
        this.definitionBuilder = new DefinitionBuilder(this.typeNameProvider);
        this.indentString = '  ';
        this.maxCommentWidth = options.maxCommentWidth || 100;
        this.commentWriter = new JavaCommentWriter(writer, this.maxCommentWidth);
    }
    writeFileContent(definition, contents) {
        this.writeLine(`package ${definition.package};`);
        contents(this);
        return this;
    }
    writeMainMethod(contents) {
        this.writeLine('public static void main(String[] args) {');
        if (contents) {
            this.increaseIndent();
            contents(this);
            this.decreaseIndent();
        }
        this.writeLine('}');
        return this;
    }
    /**
     * Writes 1 or more using directives, each on a new line.
     * @param values A collection of strings, typically namespace names.
     */
    writeUsingDirectives(...values) {
        values.forEach((v) => {
            this.writeLine(`using ${v};`);
        });
        return this;
    }
    /**
     * Writes an indented block of code, wrapped in opening and closing brackets.
     * @param contents A callback function that writes the contents.
     */
    writeCodeBlock(contents) {
        this.write(' {');
        this.writeEndOfLine();
        this.increaseIndent();
        if (contents)
            contents(this);
        this.decreaseIndent();
        this.writeLine('}');
        return this;
    }
    writeNamespaceBlock(data, contents, options) {
        if (!data)
            return this;
        const definition = elements.isPackage(data)
            ? this.definitionBuilder.buildPackageDefinition(data, options)
            : data;
        this.writeLine(`namespace ${definition.name}`);
        this.writeCodeBlock(contents);
        return this;
    }
    writeClassBlock(data, contents, options) {
        if (!data)
            return this;
        const definition = elements.isType(data)
            ? this.definitionBuilder.buildClassDefinition(data, options)
            : data;
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        this.writeIndent();
        this.writeAccessModifier(definition);
        if (definition.isStatic) {
            this.write('static ');
        }
        else if (definition.isAbstract) {
            // note that static classes cannot be abstract, hence the else-statement
            this.write('abstract ');
        }
        if (definition.isPublic) {
            this.write('public ');
        }
        this.write(`class ${definition.name}`);
        if (definition.extends) {
            this.writeExtends([definition.extends]);
        }
        this.writeImplements(definition.implements);
        this.writeCodeBlock(contents);
        return this;
    }
    writeInterfaceBlock(data, contents, options) {
        if (!data)
            return this;
        const definition = elements.isType(data)
            ? this.definitionBuilder.buildInterfaceDefinition(data, options)
            : data;
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        this.writeIndent();
        this.writeAccessModifier(definition);
        if (definition.isPublic) {
            this.write('public ');
        }
        this.write(`interface ${definition.name}`);
        this.writeExtends(definition.extends);
        this.writeCodeBlock(contents);
        return this;
    }
    writeEnumerationBlock(data, contents, options) {
        if (!data)
            return this;
        const definition = elements.isType(data)
            ? this.definitionBuilder.buildEnumDefinition(data, options)
            : data;
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        this.writeIndent();
        this.writeAccessModifier(definition);
        this.write(`enum ${definition.name}`);
        this.writeEndOfLine();
        this.writeCodeBlock(contents);
        return this;
    }
    writeEnumeration(data, options) {
        if (!data)
            return this;
        const definition = elements.isType(data)
            ? this.definitionBuilder.buildEnumDefinition(data, options)
            : data;
        this.writeEnumerationBlock(definition, () => {
            if (definition.members) {
                definition.members.forEach((memberDefinition) => {
                    this.writeEnumMember(memberDefinition);
                });
            }
        });
        return this;
    }
    writeEnumMember(data, options, isLast) {
        if (!data)
            return this;
        const definition = elements.isEnumerationLiteral(data)
            ? this.definitionBuilder.buildEnumMemberDefinition(data, isLast, options)
            : data;
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        this.writeIndent();
        this.write(definition.name);
        if (definition.value != null) {
            // using '!=' on purpose
            this.write(` = ${definition.value}`);
        }
        if (!definition.isLast) {
            this.write(',');
        }
        this.writeEndOfLine();
        return this;
    }
    writeMethodDeclaration(data, options) {
        if (!data)
            return this;
        const definition = elements.isOperation(data)
            ? this.definitionBuilder.buildMethodDefinition(data, options)
            : data;
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        if (definition.parameters) {
            this.writeJavaDocParagraph(XmlDocUtility.getXmlDocLinesForInOutParameters(definition.parameters));
        }
        const xmlDocReturns = XmlDocUtility.getXmlDocLineForReturnParameter(definition);
        if (xmlDocReturns) {
            this.writeJavaDocParagraph([xmlDocReturns]);
        }
        this.writeIndent();
        if (!definition.isPublic) {
            this.writeAccessModifier(definition); // Partial methods are implicitly private
        }
        if (definition.isPublic) {
            this.write('public ');
        }
        if (definition.isStatic) {
            this.write('static ');
        }
        else if (definition.isAbstract) {
            this.write('abstract ');
        }
        // partial methods must return void, intentional trailing white space
        else
            this.write(`${definition.returnTypeName || 'void'} `); // intentional trailing white space
        this.write(`${definition.name}(`);
        if (definition.parameters) {
            this.writeParameters(definition.parameters);
        }
        this.write(')');
        if (definition.throws) {
            this.write(' throws ');
            this.write(definition.throws.join(', '));
        }
        this.write(';');
        this.writeEndOfLine();
        return this;
    }
    writeMethodBlock(data, contents, options) {
        if (!data)
            return this;
        const definition = elements.isOperation(data)
            ? this.definitionBuilder.buildMethodDefinition(data, options)
            : data;
        // Write the documentation
        if (definition.docComment) {
            this.writeDocComment(definition.docComment);
        }
        if (definition.parameters) {
            this.writeJavaDocParagraph(XmlDocUtility.getXmlDocLinesForInOutParameters(definition.parameters));
        }
        const xmlDocReturns = definition.isConstructor
            ? null
            : XmlDocUtility.getXmlDocLineForReturnParameter(definition);
        if (xmlDocReturns) {
            this.writeJavaDocParagraph([xmlDocReturns]);
        }
        // Start of the actual method
        this.writeIndent();
        // Write the access modifier
        if (definition.isPublic) {
            this.write('public ');
        }
        else {
            this.writeAccessModifier(definition);
        }
        // Writes the default keyword if it's required
        if (definition.isDefault) {
            this.write('default ');
        }
        // Write the non-access modifier
        this.writeNonAccessModifier(definition.nonAccessModifier);
        // Write the return type
        this.write(`${definition.returnTypeName || 'void'} `); // intentional trailing white space
        this.write(`${definition.name}(`);
        if (definition.parameters) {
            this.writeParameters(definition.parameters);
        }
        this.write(')');
        if (definition.throws) {
            this.write(' throws ');
            this.write(definition.throws.join(', '));
        }
        if (definition.isAbstract) {
            this.writeEndOfLine(';');
        }
        else {
            this.writeCodeBlock(contents);
        }
        return this;
    }
    /**
     * Writes a setter
     * @param param
     * @param methodName Method name without the set prefix
     */
    writeSetterMethod(param, methodName) {
        if (!methodName) {
            methodName = param.name.replace(/^.{1}/, (letter) => letter.toLocaleUpperCase());
        }
        this.writeLine(`public void set${methodName}(${param.typeName} ${param.name}) {`);
        this.increaseIndent();
        this.writeLine(`this.${param.name} = ${param.name};`);
        this.decreaseIndent();
        this.writeLine(`}`);
    }
    /**
     * Writes a setter
     * @param param
     * @param methodName Method name without the set prefix
     */
    writeGetterMethod(param, methodName) {
        if (!methodName) {
            methodName = param.name.replace(/^.{1}/, (letter) => letter.toLocaleUpperCase());
        }
        this.writeLine(`public ${param.typeName} get${methodName}() {`);
        this.increaseIndent();
        this.writeLine(`return this.${param.name};`);
        this.decreaseIndent();
        this.writeLine(`}`);
    }
    writeParameters(data, options) {
        if (!data || !data.length)
            // without at least 1 element, we cannot determine the type
            return this;
        const definitions = elements.isParameter(data[0])
            ? this.definitionBuilder.buildParameterDefinitions(data, options)
            : data;
        if (!definitions.length)
            return this;
        let i = 0;
        definitions.forEach((p) => {
            if (i > 0) {
                this.write(', ');
            }
            this.write(p.typeName);
            if (p.isNullable) {
                this.write('?');
            }
            this.write(` ${p.name}`);
            if (p.defaultValue) {
                this.write(` = ${p.defaultValue}`);
            }
            i++;
        });
        return this;
    }
    writeProperty(prop) {
        if (prop.docComment) {
            this.writeDocComment(prop.docComment);
        }
        // Start a new, indented line
        this.writeIndent();
        // Access modifier
        this.writeAccessModifier(prop);
        // Static
        if (prop.isStatic)
            this.write('static ');
        // Type
        this.write(`${prop.typeName} `);
        // Name
        this.write(prop.name);
        // Default value
        this.writePropertyValue(prop);
        // End
        this.write(';');
        this.writeEndOfLine();
        return this;
    }
    writePropertyValue(prop) {
        const value = prop.defaultValue;
        if (value === undefined)
            return this;
        if (prop.typeName == 'String') {
            this.write(` = "${value}"`);
        }
        else {
            this.write(` = ${value}`);
        }
        return this;
    }
    writeAccessModifier(visibilityKind) {
        if (!visibilityKind)
            return this;
        const accessModifier = typeof visibilityKind == 'number'
            ? DefinitionBuilder.getAccessModifierString(visibilityKind) // VisibilityKind
            : visibilityKind.accessModifier; // TypeDefinition
        if (!accessModifier)
            return this;
        this.write(accessModifier);
        this.writeWhiteSpace();
        return this;
    }
    writeNonAccessModifier(nonAccessModifier) {
        if (!nonAccessModifier)
            return this;
        this.write(nonAccessModifier);
        this.writeWhiteSpace();
        return this;
    }
    getTypeName(element) {
        if (!element)
            return null;
        return this.typeNameProvider.getTypeName(element);
    }
    // #region deprecated
    /**
     * Writes a method declaration without a body.
     * @deprecated Use the writeMethodDeclaration() function instead.
     */
    writeInterfaceMethod(operation, options) {
        console.warn('writeInterfaceMethod is deprecated. Use the writeMethodDeclaration() function instead. ');
        this.writeMethodDeclaration(operation, options);
        return this;
    }
    /**
     * Writes an indented block of code, wrapped in a method declaration and opening and closing brackets.
     * @deprecated Use the writeMethodBlock() function instead.
     */
    writeClassMethodBlock(operation, contents, options) {
        console.warn('writeClassMethodBlock is deprecated. Use the writeMethodBlock() function instead. ');
        this.writeMethodBlock(operation, contents, options);
        return this;
    }
    writeDocComment(data) {
        if (elements.isNamedElement(data)) {
            // we have no isElement check, but this will do
            data = DefinitionBuilder.buildXmlDocSummary(data);
        }
        if (!data)
            return this;
        const array = [];
        if (typeof data == 'string') {
            // string
            array.push(data);
        }
        else {
            // string[]
            array.push(...data);
        }
        if (!array.length)
            return this;
        this.writeLine('/**');
        this.commentWriter.writeCommentLines(array, ' * ');
        this.writeLine(' */');
        return this;
    }
    writeJavaDocParagraph(data) {
        if (data == null)
            return this;
        let lines;
        if (typeof data == 'string') {
            lines = [data];
        }
        else
            lines = data;
        this.writeJsDocLines(lines);
        return this;
    }
    writeJsDocLines(lines) {
        if (lines.length === 0)
            return this;
        this.writeLine('/**');
        // Split here
        lines.forEach((line) => {
            const lineLength = line ? line.length : 0;
            if (this.maxCommentWidth > 0 && lineLength > this.maxCommentWidth) {
                // See if we can split the line
                var split = CodeWriterUtility.wordWrap(line, this.maxCommentWidth);
                split.forEach((s) => {
                    this.writeLine(`* ${s}`);
                });
            }
            else
                this.writeLine(`* ${line}`);
        });
        this.writeLine('*/');
        return this;
    }
    writeDelimitedCommentParagraph(data) {
        if (typeof data == 'string') {
            this.commentWriter.writeDelimitedCommentParagraph(data);
        }
        else
            this.commentWriter.writeDelimitedCommentLines(data);
        return this;
    }
    writeDelimitedCommentLines(paragraph) {
        console.warn('writeDelimitedCommentLines is deprecated. Use the writeDelimitedCommentParagraph() function instead.');
        this.commentWriter.writeDelimitedCommentLines(paragraph);
        return this;
    }
    //#endregion Xml Docs
    joinWrite(collection, separator, getStringFunc) {
        let isFirst = true;
        collection.forEach((c) => {
            const value = getStringFunc(c);
            if (!value)
                return;
            if (isFirst) {
                isFirst = false;
            }
            else
                this.write(separator);
            this.write(value);
        });
    }
    writeExtends(ext) {
        if (!ext || !ext.length)
            return false;
        this.write(' extends ');
        this.joinWrite(ext, ', ', (name) => name);
        return true;
    }
    writeImplements(impl) {
        if (!impl || !impl.length)
            return false;
        this.write(' implements ');
        this.joinWrite(impl, ', ', (name) => name);
        return true;
    }
    /**
     *
     * @param classList The class list to imported
     */
    writeImports(classList) {
        classList.forEach((cls) => {
            this.writeLine(cls ? `import ${cls};` : '');
        });
    }
}
